-------------------------------------------
-- Author: Gregor "farhaven" Best"       --
-- Copyright 2009 Gregor Best            --
-------------------------------------------

return {
  basic_mpd        = require 'obvious.basic_mpd',
  battery          = require 'obvious.battery',
  bluetooth        = require 'obvious.bluetooth',
  clock            = require 'obvious.clock',
  cpu              = require 'obvious.cpu',
  fs_usage         = require 'obvious.fs_usage',
  io               = require 'obvious.io',
  gps              = require 'obvious.gps',
  lib              = require 'obvious.lib',
  loadavg          = require 'obvious.loadavg',
  mem              = require 'obvious.mem',
  net              = require 'obvious.net',
  popup_run_prompt = require 'obvious.popup_run_prompt',
  umts             = require 'obvious.umts',
  volume_alsa      = require 'obvious.volume_alsa',
  volume_freebsd   = require 'obvious.volume_freebsd',
  wlan             = require 'obvious.wlan',
  temp_info        = require 'obvious.temp_info',
  keymap_switch    = require 'obvious.keymap_switch',
  weather          = require 'obvious.weather',
  music            = require 'obvious.music',
}

-- vim:ft=lua:ts=2:sw=2:sts=2:tw=80:et
