------------------------------------------
-- Author: Andrei "Garoth" Thorp        --
-- Copyright 2009 Andrei "Garoth" Thorp --
------------------------------------------

return {
  hooks  = require("obvious.lib.hooks"),
  markup = require("obvious.lib.markup"),
  mpd    = require("obvious.lib.mpd"),
  widget = require("obvious.lib.widget"),
  wlan   = require("obvious.lib.wlan"),
}

-- vim:ft=lua:ts=2:sw=2:sts=2:tw=80:et
